"""
Class to extend the functionality of the default HTML5 writer of docutils.
"""

import os.path
from docutils import nodes
import docutils.writers.html5_polyglot


class Writer(docutils.writers._html_base.Writer):
    """
    Writer class for HTML5 documents.
    This class is modified from docutils html5_polyglot.Writer (Jun 2021)
    """
    default_stylesheets = []
    default_stylesheet_dirs = ['.', os.path.abspath(os.path.dirname(__file__))]

    default_template = '../quaker_theme/template.txt'
    default_template_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), default_template)

    settings_spec = (
        'Quaker HTML Options',
        None,
        (
            ('Comma separated list of stylesheet URLs. '
             'Overrides previous --stylesheet and --stylesheet-path settings.',
             ['--stylesheet'],
             {'metavar': '<URL[,URL,...]>', 'overrides': 'stylesheet_path'}),
            ('Comma separated list of stylesheet paths.',
             ['--stylesheet-path'],
             {'metavar': '<file[,file,...]>', 'overrides': 'stylesheet',
              'default': default_stylesheets}),
            ('Embed the stylesheet(s) in the output HTML file.',
             ['--embed-stylesheet'],
             {'default': 1, 'action': 'store_true'}),
            ('Link to the stylesheet(s) in the output HTML file.',
             ['--link-stylesheet'],
             {'dest': 'embed_stylesheet', 'action': 'store_false'}),
            ('List of directories where stylesheets are found.',
             ['--stylesheet-dirs'],
             {'metavar': '<dir[,dir,...]>',
              'default': default_stylesheet_dirs}),
            ('Specify the initial header level.',
             ['--initial-header-level'],
             {'choices': '1 2 3 4 5 6'.split(), 'default': '2',
              'metavar': '<level>'}),
            ('Format for footnote references',
             ['--footnote-references'],
             {'choices': ['superscript', 'brackets'], 'default': 'brackets',
              'metavar': '<format>',
              'overrides': 'trim_footnote_reference_space'}),
            ('Format for block quote attributions',
             ['--attribution'],
             {'choices': ['dash', 'parentheses', 'parens', 'none'],
              'default': 'dash', 'metavar': '<format>'}),
            ('Remove extra vertical whitespace between items of bullet '
             'lists and enumerated lists.',
             ['--compact-lists'],
             {'default': True, 'action': 'store_true'}),
            ('Disable compact simple bullet and enumerated lists.',
             ['--no-compact-lists'],
             {'dest': 'compact_lists', 'action': 'store_false'}),
            ('Remove extra vertical whitespace between items of simple field '
             'lists.',
             ['--compact-field-lists'],
             {'default': True, 'action': 'store_true'}),
            ('Disable compact simple field lists.',
             ['--no-compact-field-lists'],
             {'dest': 'compact_field_lists', 'action': 'store_false'}),
            ('Embed images in the output HTML file, if the image '
             'files are accessible during processing.',
             ['--embed-images'],
             {'default': 0, 'action': 'store_true'}),
            ('Link to images in the output HTML file.',
             ['--link-images'],
             {'dest': 'embed_images', 'action': 'store_false'}),
            ('Added to standard table classes. '
             'Defined styles: borderless, booktabs, '
             'align-left, align-center, align-right, colwidths-auto.',
             ['--table-style'],
             {'default': ''}),
            ('Math output format (one of "MathML", "HTML", "MathJax", '
             'or "LaTeX") and option(s).',
             ['--math-output'],
             {'default': 'HTML ../quaker_theme/static/css/math.css'}),
            ('Prepend an XML declaration.',
             ['--xml-declaration'],
             {'default': False, 'action': 'store_true'}),
            ('Obfuscate email addresses to confuse harvesters.',
             ['--cloak-email-addresses'],
             {'default': False, 'action': 'store_true'}),
        ))

    config_section = 'html5 writer'

    visitor_attributes = (
        'head_prefix', 'head', 'stylesheet', 'body_prefix',
        'body_pre_docinfo', 'docinfo', 'body', 'body_suffix',
        'title', 'subtitle', 'header', 'footer', 'meta', 'fragment',
        'html_prolog', 'html_head', 'html_title', 'html_subtitle',
        'html_body', 'navigation', 'logo')

    def __init__(self):
        """
        Initialize Writer class.
        """
        super().__init__()
        self.translator_class = HTMLTranslator


class HTMLTranslator(docutils.writers.html5_polyglot.HTMLTranslator):
    """
    Class used for translating internal docutils-nodes to HTML.
    """
    def __init__(self, document):
        """
        Initialize the HTML translator
        """
        super().__init__(document)
        self.settings = document.settings

        # Set base path for every document.
        self.head.append(f'<base href="{self.settings.rel_base}">')

        # Add favicon to pages.
        if self.settings.favicon is not None:
            self.head.append('<link rel="icon" '
                             f'href="{self.settings.favicon}">')

        # Build navigation bar.
        # self.navigation = ''
        # for toc in self.settings.toc:
        #     self.navigation += directives.sphinx.TocTree.to_html(toc)
        self.navigation = '<div id="navigation-tree"></div>'

        # Add logo to pages.
        self.logo = ''
        if self.settings.logo is not None:
            self.logo = f'<img src="{self.settings.logo}" alt="Logo">'

        link = 'https://quakerdocs.nl/'

        self.footer.append(
            f'<p>&copy {self.settings.copyright}.</p>'
            '<p>Generated with <span style="color: red"> '
            '<i class="fas fa-heart"></span></i> & '
            '<span style="color: #ffcc4d"><i class="fas fa-beer"></i></span> '
            f'<a href="{link}"> QuakerDocs</a></p>')

    def visit_metadata(self, node: nodes.Element):
        """
        Skip rendering of metadata data-element.
        """
        raise nodes.SkipNode

    def visit_toc_data(self, node: nodes.Element):
        """
        Render the Table of Contents data-element.
        """
        if node['hidden']:
            raise nodes.SkipNode
            return

        for i in node.create_html(self.settings.id_map):
            self.body.append(i)

    def depart_toc_data(self, node: nodes.Element):
        pass

    def visit_pending_xref(self, node: nodes.Element):
        """
        ...
        """
        raise nodes.SkipNode

    def visit_ref_element(self, node: nodes.Element):
        """
        Find reference belonging to this element.
        """
        title = node['title']
        if node['ref'] in self.settings.id_map:
            ref = self.settings.id_map[node['ref']]

            # Make sure the title exists.
            if title is None:
                title = ref if ref.title is None else ref.title

            self.body.append(f'<a href="{ref.url}">{title}</a>')
        else:
            if title is not None:
                self.body.append(title)

    def depart_ref_element(self, node: nodes.Element):
        """
        End of rendering reference.
        """

    def visit_kbd_element(self, node: nodes.Element):
        """
        Begin rendering of keystroke element.
        """
        self.body.append('<kbd>')
        self.body.append('+'.join(f'<kbd>{key}</kbd>' for key in node['keys']))

    def depart_kbd_element(self, node: nodes.Element):
        """
        End of rendering keystroke element.
        """
        self.body.append('</kbd>')

    def visit_iframe_node(self, node: nodes.Element) -> None:
        """
        Do nothing on visit.
        """

    def depart_iframe_node(self, node: nodes.Element) -> None:
        """
        Add iframe HTML with correct attributes.
        """
        url = node['url']
        width = node['width']
        height = node['height']

        code = (f'<iframe src="{url}" width="{width}" height="{height}" '
                'frameborder="0" allow="autoplay"></iframe>')
        self.body.append(code)

    def visit_section(self, node: nodes.Element) -> None:
        """
        Adds a span before the section with the same id as the section.
        """
        if node['ids']:
            section_id = node['ids'][0]
            self.body.append(f'<span class="anchor" id="{section_id}"></span>')
        super().visit_section(node)

    def depart_title(self, node: nodes.Element) -> None:
        """
        Append bookmark button to title element.
        """
        if len(self.context) > 0:
            close_tag = self.context[-1]

            if close_tag.startswith('</h'):
                self.add_bookmark_btn(node)
        super().depart_title(node)

    def add_bookmark_btn(self, node: nodes.Element):
        """
        Add bookmark button for node to the document body.
        """
        title = node.astext()
        bookmark_id = self.create_bookmark_id(node)
        onclick = f"bookmarkClick('{bookmark_id}')"
        html = (f'<button id="{bookmark_id}" class="bookmark-btn" '
                f'onclick="{onclick}" title="{title}" value=0>'
                '<span class="icon"><i class="fa fa-bookmark-o">'
                '</i></span></button>')
        self.body.append(html)

    def create_bookmark_id(self, node: nodes.Element):
        """
        Assign a unique identifier to the bookmark.
        """
        try:
            id_str = "BM_" + str(node.parent['ids'][0])
            return id_str
        except (KeyError, IndexError):
            print('Cannot make bookmark ID, because parent ID can '
                  'not be established.')
            raise



"""
Functionality related to the application.
"""

from importlib import import_module
from docutils.parsers.rst import directives, roles

from quaker_lib.util import Config


class SphinxApp:
    """
    Class to emulate a Sphinx app for using extensions.
    """
    def __init__(self):
        # Ignore for now
        self.env = Config()
        self.config = Config()
        self.callbacks = {}
        self.handlers = {
            'html': [],
            'latex': [],
            'text': []
        }

        self.source_suffix = {}
        self.source_parsers = {}

    def get_handlers(self):
        """
        Return list of handlers that should be added to the translator.
        """
        return self.handlers

    def add_directive(self, name, directive):
        """
        Add a docutils directive to the program.
        """
        directives.register_directive(name, directive)

    def add_role(self, name, role):
        """
        Add a docutils role to the program.
        """
        roles.register_local_role(name, role)

    def add_node(self, node, **kwargs):
        """
        Placeholder function for adding node types to docutils.
        Does nothing at the moment.
        """
        ...

    def add_source_suffix(self, file_ext, file_type):
        """
        Add a source suffix to the project.
        """
        self.source_suffix.update({file_ext: file_type})

    def add_source_parser(self, parser):
        """
        Add a source parser to the project.
        """
        for file_ext, file_type in self.source_suffix.items():
            if file_type in parser.supported:
                self.source_parsers.update({file_ext: parser})
                return

    def add_post_transform(self, post_transform):
        """
        Add a post transform to the project.
        """
        ...

    def add_config_value(self, name, value, map):
        """
        Add a config value to the project.
        """
        self.config[name] = value

    def connect(self, event, callback):
        """
        Add an event callback to the project.
        """
        if event not in self.callbacks:
            self.callbacks[event] = []
        self.callbacks[event].append(callback)


def setup_extension(extension, app):
    """
    Load extension module, and call the containing setup function
    """
    mod = import_module(extension)
    setup = getattr(mod, 'setup')
    setup(app)



""" Module to generate a radix prefix trie and a list of page info.

This module implements a radix prefix trie class that only contains methods
to add new words to the trie or to store the contents of the trie in a binary
representation. A class is also implemented to find the minimum sized data type
that is able to store a given integer using the struct library. Lastly there
is also a class that can parse a block of text and create a header file
containing the resulting trie and page infos.

Notes
-----
When this module is loaded it checks wheter or not the english stopwords are
available on the machine and downloads them if they are not available.

Attributes
----------
stopwords : set
    A set of stopwords used in the english language and should be removed.

"""

import os
import re
import math
import struct
import shutil
from pathlib import Path
from typing import Tuple
from jinja2 import Template
from collections import deque
from collections import Counter
from types import SimpleNamespace


# A set of stopwords that should be ignored when part of a search.
stopwords = {
    'him', 'then', 'i', 'couldn', 'too', 'shan', 'a', 'them', 'doesnt', 'it',
    'shouldnt', 'no', 'only', 'that', 'yourself', 'because', 'ain', 'very',
    'werent', 'at', 'nor', 'below', 'mightn', 'to', 'don', 'there', 'about',
    'did', 'will', 'mustnt', 'these', 'your', 'are', 'hasn', 'from', 'what',
    'in', 'isnt', 'she', 'they', 'having', 'mightnt', 'll', 'shant', 'shes',
    'my', 'wouldn', 'have', 'am', 'some', 'as', 'on', 'where', 'most', 'not',
    'our', 'herself', 'both', 'should', 'myself', 'he', 'and', 'mustn',
    'hadnt', 'youll', 'been', 'be', 'during', 'arent', 'when', 'doing',
    'ourselves', 'after', 'than', 'youre', 'if', 'isn', 'wasn', 'shouldve',
    'didnt', 'until', 'd', 'his', 'between', 'out', 'didn', 'so', 'dont',
    'which', 'off', 'doesn', 'down', 'wouldnt', 've', 'why', 'ma', 'being',
    'wont', 'but', 'yours', 'youve', 'few', 'were', 'before', 'any', 'for',
    'through', 'up', 'again', 'into', 'same', 'of', 'its', 'aren', 'own',
    'haven', 're', 'under', 'thatll', 'was', 'this', 'such', 'over', 'neednt',
    's', 'once', 'himself', 'me', 'her', 'hasnt', 'all', 'itself', 'ours',
    'while', 'o', 'themselves', 'youd', 'each', 't', 'just', 'their', 'm',
    'won', 'hers', 'who', 'by', 'does', 'above', 'hadn', 'more', 'we',
    'weren', 'had', 'couldnt', 'is', 'or', 'an', 'needn', 'y', 'further',
    'now', 'you', 'how', 'against', 'those', 'can', 'here', 'has', 'theirs',
    'with', 'yourselves', 'the', 'wasnt', 'havent', 'whom', 'do', 'other',
    'shouldn'
}


def get_primitive(size: int) -> SimpleNamespace:
    """Choose the minimum required data type based on the size.

    Parameters
    ----------
    size : int
        The maximum value the primitive should be able to hold.

    Raises
    ------
    ValueError
        When there is too much data in the index.

    Returns
    -------
    SimpleNamespace
        An object containing the id, type and number of bytes.

    """
    prim = SimpleNamespace()
    prim.actual_bytes = math.ceil(math.log2(size) / 8) if size > 0 else 0

    if prim.actual_bytes <= 1:
        prim.id, prim.type, prim.bytes = 'B', 'unsigned char', 1
    elif prim.actual_bytes == 2:
        prim.id, prim.type, prim.bytes = 'H', 'unsigned short', 2
    elif prim.actual_bytes <= 4:
        prim.id, prim.type, prim.bytes = 'I', 'unsigned int', 4
    elif prim.actual_bytes <= 8:
        prim.id, prim.type, prim.bytes = 'Q', 'unsigned long long', 8
    else:
        raise ValueError('Too much data in the search index.')

    return prim


class Trie:
    """A class to represent a Radix tree/trie.

    Attributes
    ----------
    char : str
        The string connecting the node to its parent.
    end : bool
        Signifies wether node is an endpoint.
    children : dict
        A dictionary connecting chars to child nodes.
    pages : [(int, int)]
        A list of page indexes and word counts.

    """

    def __init__(self, char: str):
        """Constructor for the Trie class.

        Parameters
        ----------
        char : str
            The string connecting the node to its parent.

        """

        # Current character.
        self.char = char

        # Wether at end of word or not.
        self.end = False

        # Dict of children.
        self.children = {}

        # The pages where this word is found.
        self.pages = []

    def insert_helper(self, word: str, current: "Trie") -> Tuple[str, "Trie"]:
        """A helper method used to insert a fragment of a word into the trie.

        Parameters
        ----------
        word : str
            The remainder of the word to insert a fragment of.
        current : Trie
            The node of the trie from which to start searching.

        Returns
        -------
        str
            The remaining part of the word, can be empty.
        Trie
            The last node where a string was inserted.

        """

        # Loop over the children of the current node.
        for c_word, child in current.children.items():
            # Find how well the current node matches the word.
            match, remainder, word = self.match(c_word, word)

            # Exact match to start of word.
            if not remainder:
                # Move a layer deeper.
                return word, child

            # Not exact but still partial match.
            if match:
                # Create a new node for the partial match.
                new = Trie(match)

                # Update the char of the child and move it the new node.
                child.char = remainder
                new.children[remainder] = child
                current.children.pop(c_word)
                current.children[match] = new

                # Continue after the partial match.
                if word:
                    # Create the remaining word object below.
                    current = new
                    break
                # The word fits in the new node.
                return "", new

        # If none of the children matched add the remainder as a child.
        new = Trie(word)
        current.children[word] = new
        return "", new

    def insert(self, word: str, page: int, count: int):
        """Insert method to insert a new word into the trie.

        Parameters
        ----------
        word : str
            The word to be inserted.
        page : int
            The index of the page the word is in.
        count : int
            The number of times the word is found.

        """
        if not word:
            return

        # Start at the root and continue until the end of the word.
        current = self
        while word:
            word, current = self.insert_helper(word, current)

        # If the loop exited it means that the current node is an end node.
        current.end = True
        current.pages.append((page, count))

    def insert_ignore(self, word: str):
        """Insert method to insert a new word into the trie, but ignore the
        pages associated with it.

        Parameters
        ----------
        word : str
            The word to be inserted.

        """
        # Start at the root and continue until the end of the word.
        current = self
        while word:
            word, current = self.insert_helper(word, current)

        # If the loop exited it means that the current node is an end node.
        current.end = True
        current.pages = []

    @staticmethod
    def match(n_word: str, s_word: str) -> Tuple[str, str, str]:
        """Method to match a part of a word to a part in the node.

        Parameters
        ----------
        n_word : str
            The word fragment inside the node.
        s_word : str
            The fragment of the word being searched.

        Returns
        -------
        str
            The matching part of the words.
        str
            The remaining part of n_word.
        str
            The remaining part of s_word.

        """
        # In case either string is empty.
        i = -1

        # Loop over both words.
        for i, (n_char, s_char) in enumerate(zip(n_word, s_word)):
            # Stop on first difference.
            if n_char != s_char:
                break
        else:
            # Perfect match: split on end of string.
            i += 1

        # Return matching part and both remainders.
        return n_word[:i], n_word[i:], s_word[i:]

    def get_word(self, word: str):
        """"Find the node for a specific word in the trie.

        Parameters
        ----------
        word : str
            The word to search for.

        Returns
        -------
        Trie
            The node containing the data for the specified word.

        Raises
        ------
        KeyError
            When the trie does not contain the word.
        """
        current = self
        while word:
            # Loop over the children of the current node.
            for c_word, child in current.children.items():
                # Find how well the current node matches the word.
                match, remainder, word = self.match(c_word, word)

                if not remainder:
                    current = child
                    break
            else:
                # No suitable child found.
                raise KeyError

        if not current.end:
            # requested string is part of a word, but not a word itself.
            raise KeyError

        return current

    def flatten_data(self) -> SimpleNamespace:
        """ Convert the trie to an object containing arrays instead of a tree.

        Returns
        -------
        SimpleNamespace
            An object containing the data from the trie as separate arrays.

        """
        data, nodes = SimpleNamespace(), []
        data.char_p, data.page_p, data.child_s = 0, 1, 0
        data.page_s, data.page_i, data.page_c = 0, 0, 0

        # Iterate over all the nodes.
        queue = deque([[self]])
        while queue:
            # Append the node to the node list.
            node = queue[0].pop()
            node.i = len(nodes)
            nodes.append(node)
            if not queue[0]:
                queue.popleft()

            # Convert the children dictionary to a list, and traverse it later.
            node.children_list = list(node.children.values())
            if node.children_list:
                queue.append(list(reversed(node.children_list)))

            # Count the number of children, pages and characters.
            data.page_p += len(node.pages)
            data.char_p += len(node.char) + 1  # + 1 for '\0'.

            # Get the highest number of children / pages.
            data.child_s = max(data.child_s, len(node.children))
            data.page_s = max(data.page_s, len(node.pages))

            # Get the highest page index / page count in a node.
            for index, count in node.pages:
                data.page_i = max(data.page_i, index)
                data.page_c = max(data.page_c, count)

        # Get the minimum primitive data type for each of the values.
        for key, val in data.__dict__.items():
            setattr(data, key, get_primitive(val))

        # Get the smallest size primitives which can store the found values.
        data.nodes, data.node_p = nodes, get_primitive(len(nodes))
        data.ignore_node_value = (1 << data.page_p.bytes) - 1

        return data

    def to_binary(self):
        """Translate the Trie to a binary format

        Returns
        -------
        Dict
            The binary representations of the radix trie.

        """
        # Put all the nodes in a list.
        t = self.flatten_data()

        # Convert the radix trie to byte data.
        node_arr, page_arr, char_arr = [], [], []
        char_len = 0

        # Add each node to the binary arrays.
        for node in t.nodes:
            struct_format = (t.char_p.id + t.node_p.id + t.page_p.id
                             + t.child_s.id + t.page_s.id)

            # Check if this is a word we should ignore.
            pages_ptr = len(page_arr) // (t.page_i.bytes + t.page_c.bytes)
            if not node.pages and node.end:
                pages_ptr = t.ignore_node_value

            # Add the node data.
            node_arr += struct.pack(
                struct_format,
                char_len,
                node.children_list[0].i if node.children else 0,
                pages_ptr,
                len(node.children), len(node.pages)
            )

            # Add the pages.
            for page in node.pages:
                page_arr += struct.pack(t.page_i.id + t.page_c.id, *page)

            # Add the characters.
            char_arr += ['"'] + list(node.char) + ['\\0"']
            char_len += len(node.char) + 1

        # Convert the data to C arrays.
        t.node_arr = ','.join(map(str, node_arr))
        t.page_arr = ','.join(map(str, page_arr))
        t.char_arr = ''.join(char_arr)

        return t


class IndexGenerator:
    """
    A class to generate the indexing/trie used for searching as well as the
    translation of page indices to page info.

    Parameters
    ----------
    title_weight : int, optional
        How often words in the title are counted (the default is 5).

    Attributes
    ----------
    urltitles : [(str, str)]
        A list of urls and titles of the pages.
    trie : Trie
        A radix trie to store the words.
    remover
        Regex to match only letters and numbers.

    """

    def __init__(self):
        """Constructor for the IndexGenerator class."""
        self.urltitles = []  # [(url, title), ...]
        self.trie = Trie("")  # root of the prefix trie
        self.remover = re.compile('[^\\w\\s\\n]')

    def add_file(self, content: list, title: str, url: str,
                 priority: float = 1.0, t_weight: int = 5):
        """Add a file to the index.

        Parameters
        ----------
        content : list
            A list of words from the file.
        title : str
            The title of the document.
        url : str
            The url of the page.

        """
        # Change to lowercase, separate _ and only keep letters/numbers.
        content += (title + ' ') * t_weight
        content = content.lower().replace('_', ' ').replace('.', ' ')
        content = self.remover.sub('', content)

        # Remove stopwords.
        content = [word for word in content.split()]

        # Count occurrences of words in page.
        word_counter = Counter(content)

        # Get the index of the current page and update the list.
        i = len(self.urltitles)
        self.urltitles.append((url, title))

        # Add the words from this page to the trie.
        for word, count in sorted(word_counter.items(), key=lambda x: x[1]):
            self.trie.insert(word, i, int(count * priority))

    def build(self, temp_path: Path, dest_path: Path):
        """Write the search index file to the destination directory.

        Parameters
        ----------
        temp_path : str
            Where to put the temporary compile output.
        dest_path : str
            Where to place the javascript and webassembly files.

        """
        print('Building the search index assembly')
        source_path = Path(__file__).parent / 'wasm'

        if not source_path.exists():
            raise FileNotFoundError('Wasm source files are not found.')

        # Insert the stopwords into the trie.
        for stopword in stopwords:
            self.trie.insert_ignore(stopword)

        # Generate the search.h.
        data = self.trie.to_binary()
        template = Template((source_path / 'search.h.jinja').read_text())

        # Write the search index to header.
        temp_path = temp_path / 'search'
        temp_path.mkdir(parents=True, exist_ok=True)
        with open(temp_path / 'search.h', 'w') as f:
            f.write('/*=== AUTOMATICALLY GENERATED FILE ===*/\n\n')
            f.write(template.render(urltitles=self.urltitles,
                                    stopwords=stopwords,
                                    **data.__dict__))

        # Compile the C code using clang.
        object_file = temp_path / 'search_data.o'
        source_file = str(source_path / 'search.c')

        cmnd = ['clang', '-Os', '-c',
                '--target=wasm32', '-march=wasm32',
                '-fno-builtin',
                '-I', f'"{temp_path}"',
                '-o', f'{object_file}',
                f'"{source_file}"']
        os.system(' '.join(cmnd))

        # Find the wasm-ld executable.
        wasm_linker = 'wasm-ld'
        if shutil.which(wasm_linker) is None:
            wasm_linker = '/usr/lib/llvm-10/bin/wasm-ld'
        if shutil.which(wasm_linker) is None:
            print('wasm-ld is not found! You should probably '
                  'install it (TODO: better message)')

        # Link the object file to create a usable webassembly binary.
        cmnd = [wasm_linker, '--no-entry',
                '-o', f"{dest_path / 'search_data.wasm'}",
                f'{object_file}']
        dest_path.mkdir(parents=True, exist_ok=True)
        os.system(' '.join(cmnd))



"""Module implementing the page class which represents a page to be parsed.

The page object can parse the contents of the loaded file, handle the logic
for linking references to url's, and can write the pages to the destination.

"""

import os
import docutils.core

from docutils import nodes
from types import SimpleNamespace
from quaker_lib.util import make_id
from quaker_lib.directives.sphinx import ref_role
from quaker_lib.directives.metadata import get_metadata


class Page:
    """Handler for a single page.

    Responsible for parsing and writing files as well as keeping track of any
    unresolved references so that it knows when writing is possible.

    Parameters
    ----------
    main : Main
        The instance of main, from which settings are extracted.
    path : pathlib.Path
        The path to the file to be handled.

    Attributes
    ----------
    main : Main
        Store the instance of main.
    path : pathlib.Path
        Store the path to be handled.
    path_html : str
        The path with the html suffix as a string.
    src : pathlib.Path
        The path to the file from the source folder ('absolute' for us).
    dest : pathlib.Path
        The path to the file from the destination folder ('absolute' for us).
    doctree
        The AST generated by docutils.
    unresolved_references : int
        The number of unresolved references, used to check when to write.

    """

    def __init__(self, main, path):
        self.main = main
        self.path = path
        self.path_html = str(path.with_suffix('.html'))
        self.src = main.source_path / path
        self.dest = main.dest_path / self.path_html
        self.doctree = None
        self.unresolved_references = 0

    def parse(self):
        """ Parse an RST file. """

        # Add prologue and epilogue to source file.
        content = (f"{self.main.conf_vars.get('rst_prolog', '')}\n"
                   f"{self.src.read_text()}\n"
                   f"{self.main.conf_vars.get('rst_epilog', '')}")

        # Give the reference role access to the current page.
        ref_role.page = self

        self.main.docutil_settings['env'].docname = self.src.name

        # Parse the file contents.
        self.doctree = docutils.core.publish_doctree(
            content,
            parser=self.main.source_parsers.get(self.src.suffix)(),
            source_path=str(self.src),
            settings_overrides={
                **self.main.docutil_settings,
                'page': self,
                'main': self.main,
            }
        )

        # Find the page title.
        try:
            self.title = next(iter(self.doctree.traverse(nodes.title)))[0]
            self.title = self.title.astext()
        except StopIteration:
            self.title = ''

        self.handle_references()

        # Write if all the references are already resolved.
        if self.unresolved_references == 0:
            self.write()

    def use_reference(self, ref: str) -> str:
        """ Handle the logic for safely linking references.

        Paramaters
        ----------
        ref : str
            The reference to be linked.

        Returns
        -------
        ref : str
            The potentially altered reference.

        """
        # If the reference is an off site url keep it.
        if (ref.count('.') >= 2
                or ref.startswith('https://')
                or ref.startswith('http://')):
            return ref

        # If absolute path simply add to waiting list.
        if ref[0] == '/':
            pass
        # If it starts with a # handle it as a local section.
        elif ref[0] == '#':
            ref = '/' + str(self.path) + ref
        # If relative path prepend the current path.
        elif '/' in ref or '.' in ref or '#' in ref:
            ref = '/' + str(self.path.parent / ref)
        # Otherwise assume (global) id (else user should prepend ./).
        else:
            pass

        ref = make_id(ref)

        # Update the unresolved references if it cannot be linked yet.
        if ref not in self.main.id_map:
            self.main.waiting[ref].append(self)
            self.unresolved_references += 1

        return ref

    def id_to_map(self, id: str, node, anchor: str = ''):
        """Helper method to add a new id to the id_map.

        Parameters
        ----------
        id : str
            The id to be added.
        node
            The node the id maps to.
        anchor : str
            The potential section anchor to append.

        Returns
        -------
        references : list
            List of the references that were added to the map.

        """
        references = [id, '/' + str(self.path.with_suffix('')) + anchor]

        # Create the linking content object.
        contents = SimpleNamespace()
        contents.url = self.path_html + anchor
        contents.title = None
        contents.sections = []  # We will fill these later.

        # Get the title.
        title_node = node.next_node(nodes.Titular)
        if title_node:
            contents.title = title_node.astext()

        # Add all references.
        for ref in references:
            self.main.id_map[ref] = contents

        return references

    def handle_references(self):
        """ Take care of the references and linkage. """

        # Add the file name as an id as well.
        page_id = make_id(str(self.path.name))

        # Add all the ids to the reference map.
        all_references = self.id_to_map(page_id, self.doctree)

        for id, node in self.doctree.ids.items():
            all_references += self.id_to_map(id, node, anchor='#' + id)

        # Loop over the sections of this node, and add these to the
        # contents of this page's reference.
        stack = [self.doctree]
        first = True
        while stack:
            node = stack.pop()

            # Skip the node if no ids were found.
            if not node.attributes['ids']:
                continue

            id = node.attributes['ids'][0]
            sections = self.main.id_map[id].sections

            if first:
                first = False
                self.main.id_map[all_references[1]].sections = sections

            for id in node.attributes['ids'][1:]:
                self.main.id_map[id].sections = sections

            # Fill the sections:
            for child in node.children:
                if not isinstance(child, nodes.section):
                    continue

                id = child.attributes['ids'][0]
                id = '/' + make_id(str(self.path) + '#' + id)
                sections.append(id)

                # Go over the subsections.
                if len(child.children) > 0:
                    stack.append(child)

        # Resolve all the references we have added.
        for ref in all_references:
            try:
                for page in self.main.waiting.pop(ref):
                    page.unresolved_references -= 1
                    if page.unresolved_references == 0:
                        page.write()
            except KeyError:
                pass

    def get_settings_overrides(self):
        """Get the settings_override for this page.
        Used as arg for the call to docutils.core.publish_from_doctree
        in write.

        Returns
        -------
        dict
            The argument expected by publish_from_doctree.
        """
        return {
            'toc': self.main.toc_navigation,
            'template': self.main.theme.get_template(),
            'stylesheet': os.path.join(
                '_static',
                self.main.conf_vars.get('html_style',
                                        self.main.theme.get_style())),
            'src_dir': self.main.source_path,
            'dest_dir': self.main.relative_path(self.src),
            'html_path': self.path_html,
            'embed_stylesheet': False,
            'rel_base': os.path.relpath(self.main.dest_path,
                                        self.dest.parent),
            'handlers': self.main.sp_app.get_handlers(),
            'favicon': self.main.conf_vars.get('html_favicon', None),
            'logo': self.main.conf_vars.get('html_logo', None),
            'copyright': self.main.conf_vars.get('copyright', ''),
            'id_map': self.main.id_map
        }

    def del_skipped_nodes(self):
        """Delete the nodes from the page's doctree that match SKIP_TAGS."""
        for node in self.doctree.traverse():
            for i, child in reversed(list(enumerate(node.children))):
                if child.tagname in self.main.SKIP_TAGS:
                    del node[i]

    def write(self):
        """
        Parse an RST file and write its contents to a file.
        """

        # Get the page metadata.
        metadata = get_metadata(self.doctree)
        if metadata.ignore:
            return

        # Delete the nodes we want to skip.
        self.del_skipped_nodes()

        # Collect all the text content to add the page to the index.
        content = ' '.join(n.astext()
                           for n in self.doctree.traverse(nodes.Text))
        self.main.idx.add_file(content, self.title,
                               self.path_html, metadata.priority)

        # Create the output file contents.
        output = docutils.core.publish_from_doctree(
            self.doctree,
            destination_path=self.dest,
            writer=self.main.writer(),
            settings_overrides=self.get_settings_overrides())

        # Write the document to a file.
        self.dest.parent.mkdir(parents=True, exist_ok=True)
        self.dest.write_bytes(output)



"""
Implement directives used in the CodeGrade reStructuredText.
"""

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives


class ExampleDirective(Directive):
    """
    Directive for an admonition with title "Example"
    """
    has_content = True
    optional_arguments = 100

    def run(self):
        """
        Create nodes for directive
        """
        wrapper = nodes.admonition('', classes=['example'])
        title = "Example"
        # Add title to example text if it is supplied.
        if len(self.arguments) > 0:
            title += ": " + ' '.join(self.arguments)

        wrapper.append(nodes.paragraph('', title, classes=['example-title']))

        # Content
        content_wrapper = nodes.compound(classes=['example-content'])
        wrapper.append(content_wrapper)

        self.state.nested_parse(
            self.content, self.content_offset, content_wrapper
        )

        return [wrapper]


class iframe_node(nodes.General, nodes.Element):
    """
    Container class for IFrameDirective.
    """
    ...


class IFrameDirective(Directive):
    """
    Directive for HTML iframes.
    """
    option_spec = {
        'width': int,
        'height': int
    }

    required_arguments = 1
    optional_arguments = 2

    def run(self):
        """
        Create nodes for this directive
        """
        if self.arguments:
            url = self.arguments[0]
            node = iframe_node()
            xy_ratio = 3/5

            if self.options:
                width = self.options['width']
                height = self.options['height'] if len(self.options) > 1 else \
                    width * xy_ratio
            else:
                width = 500
                height = width * xy_ratio

            node['url'] = url
            node['width'] = int(width)
            node['height'] = int(height)

        return [node]


def setup():
    """
    Setup function for this module.
    """
    directives.register_directive('example', ExampleDirective)
    directives.register_directive('iframe', IFrameDirective)



"""
Implement directives used by Sphinx

All implementations of the directives have been inspired by:
`https://github.com/sphinx-doc/sphinx/blob/4.x/sphinx/directives/other.py`
"""

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives, roles
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst.directives.misc import Class, Include
from importlib import import_module
import inspect

from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter

from quaker_lib import util


class kbd_element(nodes.General, nodes.Element):
    """
    Empty node for rendering keyboard inputs
    """
    pass


class toc_data(nodes.General, nodes.Element):
    """Node representing the ToC data.

    Stores the configuration options of the given ToC as well as the references
    given. It is also responsible for it's own html translation.

    """
    def create_html(self, id_map: dict, css_class: str = 'toc'):
        """Turn the contents of the ToC into a html table.

        Parameters
        ----------
        id_map : dict
            Dictionary mapping the references to content objects.
        css_class : str
            Specify which css class the toc uses.

        Returns
        -------
        body : list
            List of html strings.

        """
        # Make max_depth and collapse depth complete if -1.
        max_depth = self['maxdepth']
        if max_depth < 0:
            max_depth = 10000

        collapse_depth = self['collapsedepth']
        if collapse_depth < 0:
            collapse_depth = max_depth

        list_tag = 'ol' if self['numbered'] else 'ul'

        body = [f'<p class="caption {css_class}-label">\n'
                '\t<span class="caption-text">\n'
                f'\t{self["caption"]}\n\t\t</span>\n\t</p>'
                f'<{list_tag}{self.get_numbered_type(0)} '
                f'class="{css_class}-list">\n']

        stack = [list(reversed(self['entries']))]
        while stack:
            depth = len(stack)
            if not stack[-1]:
                stack.pop()
                continue

            current = stack[-1].pop()
            tab_size = depth * '\t'

            if current is None:
                # Print end of the sublist.
                body.append(tab_size)
                body.append(f'</{list_tag}>\n')
                body.append('</li>\n')
                continue

            collapsed = ' is-collapsed' if depth > collapse_depth else ''

            title, ref_id = current
            body.append('<li><span class="level mb-0"><a ')

            ref = id_map.get(ref_id, None)
            if ref is not None:
                if title is None:
                    title = ref.url if ref.title is None else ref.title

                if '#' in ref.url:
                    body.append(f'onClick="expandSidebar(\'{ref.url}\')" ')

                body.append(f'href="{ref.url}">{title}</a>')

                if len(ref.sections) > 0 and depth < max_depth:
                    if collapsed:
                        body.append('<span onclick='
                                    '"toggleExpand(this.parentNode)"'
                                    ' class="is-clickable icon is-small '
                                    'level-right">'
                                    '<i class="fa arrow-icon fa-angle-right" '
                                    'aria-hidden="true"></i></span>')

                    body.append('</span>')
                    body.append(f'<{list_tag}{self.get_numbered_type(depth)} '
                                f'class="{css_class}-list{collapsed}">\n')

                    new = [None] + [(None, sec)
                                    for sec in reversed(ref.sections)]
                    stack.append(new)
                else:
                    body.append('</span></li>\n')
            else:
                if title is None:
                    title = ref_id
                body.append(f'href="{ref_id}">{title}</a></span></li>\n')

        body.append(f'</{list_tag}>\n')
        return body

    def get_numbered_type(self, depth: int):
        """ Helper method to get the type of number labels for enumerates. """
        if self['numbered']:
            nt = self['number_types']
            return f' type="{nt[depth % len(nt)]}"'
        else:
            return ''


class ref_element(nodes.General, nodes.Element):
    """
    Custom reference node to handle unparsed pages.
    """
    pass


class Only(Directive):
    """
    Directive for only including content for certain builds.
    """
    required_arguments = 1
    optional_arguments = 1

    has_content = True
    option_spec = {}

    def run(self):
        """
        Create nodes for this directive.
        """
        node = nodes.container()

        if (self.arguments[0] == 'html'
           or (self.arguments[0] == 'not' and self.arguments[1] != 'html')):
            self.state.nested_parse(self.content, self.content_offset, node,
                                    match_titles=True)
            return [*node]

        return []


class TocTree(Directive):
    """Directive for generating a Table of Contents

    This also includes the generation of the navbar as this uses the ToC
    outlined in the master file of the website.

    Attributes
    ----------
    has_content : bool
        Tell docutils that this directive uses it's content.
    option_spec : dict
        Tell docutils which options are needed and their types.

    """
    has_content = True

    option_spec = {
        'maxdepth': int,
        'collapsedepth': int,
        'name': directives.unchanged,
        'caption': directives.unchanged_required,
        'numbered': directives.flag,
        'number_types': directives.unchanged,
        'hidden': directives.flag,
        # Ignore:
        'glob': directives.flag,
        'includehidden': directives.flag,
        'titlesonly': directives.flag,
        'reversed': directives.flag,
    }

    def run(self):
        """Code that is being run for the directive.

        Returns
        -------
        [tocdata] : list
            A node containing the information necessary for the ToC.

        """
        tocdata = toc_data()
        tocdata['maxdepth'] = self.options.get('maxdepth', -1)
        tocdata['collapsedepth'] = self.options.get('collapsedepth', 1)
        tocdata['caption'] = self.options.get('caption', '')
        tocdata['reversed'] = 'reversed' in self.options
        tocdata['numbered'] = 'numbered' in self.options
        tocdata['number_types'] = self.options.get('number_types', '1 a i A I')
        tocdata['number_types'] = tocdata['number_types'].split()
        tocdata['hidden'] = 'hidden' in self.options

        tocdata['entries'] = self.parse_content()
        return [tocdata]

    def parse_content(self):
        """ Fill the toctree data structure with entries. """
        result = []
        # Add each entry to the list as a title and reference tuple.
        for entry in self.content:
            # Check if current entry is in format 'Some Title <some_link>'.
            explicit_link = util.link_explicit(entry)

            if explicit_link:
                title, ref = explicit_link
            else:
                title = None
                ref = entry

            # Get the actual reference from the id_map using the appropriate
            # logic safely.
            ref = self.state.document.settings.page.use_reference(ref)
            result.append((title, ref))
        return result


class CodeBlock(Directive):
    """Directive for displaying code samples.

    Attributes
    ----------
    has_content : bool
        Tell docutils that this directive uses it's content.
    option_spec : dict
        Tell docutils which options are needed and their types.
    optional_arguments : int
        Tell docutils how many optional arguments to expect.
    """
    option_spec = {
        'name': directives.unchanged,
        'linenos': directives.flag,
        'lineno-start': int,
        'caption': directives.unchanged_required
    }

    has_content = True
    optional_arguments = 1

    def run(self):
        """ Create nodes for this directive. """
        language = self.arguments[0] if len(self.arguments) > 0 else None
        linenos = 'linenos' in self.options
        linenostart = self.options.get('lineno-start', 1)
        caption = self.options.get('caption', '')
        code = "\n".join(self.content)

        formatter = HtmlFormatter(linenos=linenos, linenostart=linenostart)
        if language is not None:
            lexer = get_lexer_by_name(language, stripall=True)
            code = highlight(code, lexer, formatter)
        else:
            code = '<div class="highlight"><pre>%s</pre></div>' % code

        wrappernode = nodes.compound(classes=[f"highlight {language}"])
        wrappernode.append(nodes.raw('', code, format="html"))
        wrappernode.append(nodes.paragraph('', caption))

        return [wrappernode]


class AutoModule(Directive):
    """
    Directive for automatically documenting a Python module.
    """
    required_arguments = 1
    option_spec = {
        'members': directives.flag,
        'show-inheritance': directives.flag,
        'undoc-members': directives.flag
    }

    class_translation = {
        'type': 'class',
        'function': 'def'
    }

    def run(self):
        """
        Create nodes for this directive.
        """
        modname = self.arguments[0]
        mod = import_module(modname)

        ret = nodes.compound(classes=['automodule'])
        ret.append(nodes.emphasis('', modname))
        ret.append(nodes.term('', mod.__doc__))
        if 'members' in self.options:
            members = [(i, getattr(mod, i)) for i in dir(mod)
                       if not i.startswith('__')
                       and not inspect.ismodule(getattr(mod, i))]
            members = [(m, n) for m, n in members
                       if hasattr(n, '__module__') and n.__module__ == modname]

            for member in members:
                node = nodes.definition_list()
                mem_type = type(member[1]).__name__

                if mem_type == 'type':
                    init_params = inspect.getargspec(member[1].__init__).args
                    function_params = f'({", ".join(init_params)})'
                elif mem_type == 'function':
                    params = inspect.getargspec(member[1]).args
                    function_params = f'({", ".join(params)})'

                type_name = self.class_translation.get(mem_type, mem_type)
                items = [
                    nodes.emphasis('', f'{type_name}', classes=['type_name']),
                    nodes.emphasis('', f'{modname}.'),
                    nodes.emphasis('', member[0]),
                    nodes.emphasis('', function_params),
                    nodes.term('', member[1].__doc__)
                ]

                # TODO: Display member attributes/methods?

                node.append(nodes.definition_list_item('', *items))
                ret.append(node)

        return [ret]


def ref_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Role for creating hyperlink to other documents.
    """
    explicit_link = util.link_explicit(text)

    if explicit_link is not None:
        title, ref = explicit_link
    else:
        ref, title = text, None

    node = ref_element()
    node['title'] = title
    node['ref'] = ref_role.page.use_reference(ref)

    return [node], []


def kbd_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Role for displaying keyboard inputs.
    """
    set_classes(options)
    node = kbd_element()
    node['keys'] = text.split('+')

    return [node], []


def setup():
    """
    Setup function for this 'extension'
    """
    directives.register_directive('only', Only)
    directives.register_directive('rst-class', Class)
    directives.register_directive('include', Include)
    directives.register_directive('toctree', TocTree)
    directives.register_directive('code-block', CodeBlock)
    directives.register_directive('automodule', AutoModule)

    roles.register_canonical_role('ref', ref_role)
    roles.register_canonical_role('kbd', kbd_role)



"""
Functionality related to theme parsing.
"""

from distutils.dir_util import copy_tree
from configparser import ConfigParser
import os.path


class Theme:
    """
    Class for theme reading.
    """
    def __init__(self, theme_name, theme_path, templ_path):
        """
        Get all data related to a class.
        """
        self.theme_path = None
        self.inherit = None
        self.stylesheet = None

        # Get the paths to places where themes could be.
        src_file_path = os.path.abspath(os.path.dirname(__file__))
        dirs = [os.path.join(src_file_path, 'static')]

        if theme_path is not None:
            for cur_dir in theme_path:
                dirs.append(cur_dir)

        # Find the requested theme in the given directories.
        for current_dir in dirs:
            path = os.path.join(current_dir, theme_name)
            if os.path.exists(path) and os.path.isdir(path):
                self.theme_path = path
                break
        if self.theme_path is None:
            raise RuntimeError(f'Failed to find theme {theme_name}!')

        # Save additional paths to templates
        self.templates_path = [os.path.abspath(path) for path in templ_path] \
            if templ_path is not None else []

        # Parse the theme configuration.
        config = ConfigParser()
        config.read(os.path.join(self.theme_path, 'theme.conf'))

        inherit_name = config.get('theme', 'inherit', fallback=None)
        if inherit_name is not None:
            self.inherit = Theme(inherit_name, theme_path, [])
        self.stylesheet = config.get('theme', 'stylesheet', fallback=None)

    def get_template(self, filename='template.txt'):
        """
        Return the path to the template.
        """
        for current_dir in self.templates_path:
            path = os.path.join(current_dir, filename)
            if os.path.exists(path) and os.path.isfile(path):
                return path

        return self.get_file(filename)

    def get_file(self, filename):
        """
        Return the path to a file in the theme.
        """
        path = os.path.join(self.theme_path, filename)
        if os.path.exists(path) and os.path.isfile(path):
            return path
        if self.inherit is not None:
            return self.inherit.get_file(filename)

        return None

    def get_style(self):
        """
        Return the path to the stylesheet for this theme.
        """
        if self.stylesheet is not None:
            return self.stylesheet
        if self.inherit is not None:
            return self.inherit.get_style()

        return None

    def copy_files(self, dest_path):
        """
        Copy files from the theme to the static folder in the build directory.
        """
        if self.inherit is not None:
            self.inherit.copy_files(dest_path)

        if os.path.exists(os.path.join(self.theme_path, 'static')):
            copy_tree(os.path.join(self.theme_path, 'static'),
                      os.path.join(dest_path, '_static'),
                      update=1)



"""Module that manages the main logic for QuakerDocs

This module implements a Main class that can be used to translate Rst or
markdown documentation to a website that is both static and interactive.
The resulting website does not need a database or any server side logic, it
is completely client-side yet does include a fast live search, an interactive
sidebar, bookmarks and custom html or iframes are also supported.

This should be a drop in replacement for sphynx in most cases and supports
most of it's extensions. Custom extensions, directives and roles should be
supported as well.

"""

import os
import pkgutil
import fnmatch
import importlib
from pathlib import Path
from collections import defaultdict

from distutils.dir_util import copy_tree

from quaker_lib import index
from quaker_lib import directives
from quaker_lib import application
from quaker_lib import html5writer

from quaker_lib.page import Page
from quaker_lib.theme import Theme

from docutils.parsers.rst import Parser
from argparse import Namespace


class Main:
    """The central component of the program, from which everything else is called.

    Parameters
    ----------
    source_path : pathlib.Path
        The directory containing the source files.
    dest_path : pathlib.Path
        The directory to write the output.
    builder : str
        The builder used for the generator.

    """
    SKIP_TAGS = {'system_message', 'problematic'}

    # Mapping builder name to (output file extension, writer class) tuple
    supported_builders = {
        'html': ('html', html5writer.Writer)
    }

    def __init__(self, source_path: Path, dest_path: Path, builder: Path):
        # Create and store the various paths.
        self.source_path = source_path
        self.dest_path = dest_path
        self.temp_path = dest_path.parent / 'tmp' / dest_path.name
        self.static_dest_path = dest_path / '_static'
        self.script_dest_path = self.static_dest_path / 'js'

        # Store the builder and the parser.
        self.builder = builder
        self.source_parsers = {'.rst': Parser}

        self.sp_app = None
        self.theme = None
        self.file_ext = '.out'
        self.writer = None
        self.idx = None

        # List to store the nodes for the navbar and id_map to link references.
        self.toc_navigation = []
        self.id_map = {}

        # Import and setup all directives.
        dir_path = Path(__file__).parent / 'directives'
        for _, module, _ in pkgutil.iter_modules([str(dir_path)]):
            module = importlib.import_module(f'quaker_lib.directives.{module}')
            module.setup()

        # A waiting queue to handle any linking.
        self.waiting = defaultdict(list)

        # The general settings for parsing using docutils.
        self.docutil_settings = {
            'src_dir': self.source_path,
            'dst_dir': self.dest_path,
        }

        self.conf_vars = {
            'exclude_patterns': [],
            'html_static_path': [],
        }

    def relative_path(self, path: Path) -> Path:
        """ Get the path of a source directory relative to the source file. """
        return Path(path).relative_to(self.source_path)

    def read_conf(self):
        """
        Read conf.py from the source directory and save the configuration.
        """
        global_vars = {
            '__file__': 'conf.py',
            '__name__': '__main__'
        }
        conf_vars = {}

        conf_path = Path('conf.py')

        # Check if the file exists.
        if not conf_path.exists():
            print("Error: conf.py not found in source directory.")
            exit(1)

        with conf_path.open('r') as conf_file:
            exec(conf_file.read(), global_vars, conf_vars)
        self.conf_vars.update(conf_vars)

        # Make sure source_suffix is a list.
        suffix = self.conf_vars['source_suffix']
        if suffix is None:
            self.conf_vars['source_suffix'] = ['.rst']
        elif isinstance(suffix, str):
            self.conf_vars['source_suffix'] = [suffix]

        # Exclude static files, as they should not be processed.
        self.conf_vars['exclude_patterns'] += \
            self.conf_vars.get('html_static_path', [])
        self.conf_vars['exclude_patterns'] += \
            self.conf_vars.get('templates_path', [])

    def load_extensions(self):
        """
        Load user configuration and extensions.
        """
        prev_cwd = Path.cwd()
        os.chdir(self.source_path)

        self.read_conf()
        self.sp_app = application.SphinxApp()
        for ext in self.conf_vars['extensions']:
            application.setup_extension(ext, self.sp_app)

        # Add new source suffixes and parsers
        if len(self.sp_app.source_suffix) > 0:
            self.conf_vars['source_suffix'].append(*self.sp_app.source_suffix)
        self.source_parsers.update(self.sp_app.source_parsers)

        # Callback to extensions.
        for name, lst in self.sp_app.callbacks.items():
            if name == 'builder-inited':
                [callback(self.sp_app) for callback in lst]

        self.sp_app.env.config = Namespace(**self.sp_app.config)
        self.sp_app.env.docname = None
        self.sp_app.env.metadata = Namespace()
        self.sp_app.env.metadata.setdefault = lambda x, y: {}
        self.docutil_settings.update({
            'env': self.sp_app.env,
            'source_suffix': self.conf_vars['source_suffix']
        })

        # Get path to theme
        self.theme = Theme(self.conf_vars.get('html_theme', 'quaker_theme'),
                           self.conf_vars.get('html_theme_path'),
                           self.conf_vars.get('templates_path', []))

        os.chdir(prev_cwd)

    def generate(self):
        """ Generate the html files from the source files.

        Read all the input files from the source directory, parse them, and
        output the results to the build directory.

        Raises
        ------
        NotImplementedError
            When an unsupported builder is specified.
        """
        # Check if requested format is supported.
        if self.builder not in Main.supported_builders:
            raise NotImplementedError("Requested builder not supported!")

        self.file_ext, self.writer = Main.supported_builders[self.builder]

        # Create the destination directory if it does not exist.
        self.dest_path.mkdir(parents=True, exist_ok=True)

        # Load user configuration and extensions.
        self.load_extensions()

        # Set-up index generator and build the files.
        self.idx = index.IndexGenerator()
        self.build_files()

        # Set-up Table of Contents data and build the search index
        self.script_dest_path.mkdir(parents=True, exist_ok=True)
        self.build_global_toc()
        self.idx.build(self.temp_path, self.script_dest_path)

        # Copy the Javascript source files to the static directory in build.
        copy_tree(Path(__file__).parent / 'static' / 'js',
                  str(self.script_dest_path),
                  update=1)

        # Copy the directories from our theme directly to the dest folder.
        self.theme.copy_files(self.dest_path)

        # Copy static files from the source directly to the static dest folder.
        self.copy_static_files()

        # Delete the temporary build files.
        # TODO implement and add flag because they are needed for local search
        # and potentially other stuff for testing.

        print(f'The generated documents have been saved in {self.dest_path}')

    def build_files(self):
        """Iterate over files in source directory and save in [(path, content)]

        This method parses and translates the source files and writes the
        resulting output files to the correct destination. Ideally only one
        file is kept in memory at a time, but in order to resolve linking some
        files have to wait for others to finish first.

        """
        master_doc = self.conf_vars.get('master_doc', 'index')

        for root, _, files in os.walk(self.source_path):
            # Read, parse and write the source files to html.
            for file in files:
                path = self.relative_path(root) / file

                # Skip excluded files.
                if self.is_excluded(path):
                    continue

                # Parse the files if found (parse also writes when possible).
                if path.suffix in self.conf_vars['source_suffix']:
                    page = Page(self, path)
                    page.parse()
                else:
                    page = None

                # Add the toc trees of the master doc page for the
                # creation of the sidebar.
                if (page is not None and
                        (str(path) == master_doc
                         or str(path.with_suffix('')) == master_doc)):
                    toc_node_type = directives.sphinx.toc_data
                    for toc_node in page.doctree.traverse(toc_node_type):
                        self.toc_navigation.append(toc_node)

        # Check if all the references are resolved.
        for ref_name, pages in self.waiting.items():
            for page in pages:
                print(f'Warning: {page.src} contains an unresolved '
                      f'reference "{ref_name}"')
                page.write()

    def is_excluded(self, path: Path) -> bool:
        """
        Check whether a file is supposed to be excluded.

        Parameters
        ----------
        path : pathlib.Path
            The path to the file to be checked

        Returns
        -------
        bool
            Whether the path should be excluded (true).
        """
        exclude_pats = self.conf_vars['exclude_patterns']
        return any(fnmatch.fnmatch(path, pattern) for pattern in exclude_pats)

    def build_global_toc(self):
        """ Write the ToC's from master_doc as a navbar. """
        with (self.script_dest_path / 'load_navbar.js').open('w') as f:
            # Create javascript that inserts it so it can be included.
            f.write('document.getElementById("navigation-tree").innerHTML = `')

            # Set parameters for each node and create the html.
            for node in self.toc_navigation:
                node['numbered'] = False
                node['maxdepth'] = -1
                node['collapsedepth'] = 0
                for html in node.create_html(self.id_map, 'menu'):
                    f.write(html)
            # End the javascript.
            f.write('`;')

    def copy_static_files(self):
        """
        Copy all the files from the static path to the destination path.
        """
        for path in self.conf_vars['html_static_path']:
            copy_tree(str(self.source_path / path),
                      str(self.static_dest_path),
                      update=1)

    def init_empty_project(self):
        """
        Initializes an empty project
        """
        path = Path(self.source_path)
        qs_path = Path(__file__).parent / '..' / 'quaker_lib' / 'quickstart'

        if not path.is_dir() or not path.exists():
            copy_tree(str(qs_path), str(path), update=1)
            print('Created an empty project in', self.source_path)
        else:
            print(f'Error: directory \'{path}\' already exists!')



# -- General configuration ------------------------------------------------

# General information about the project.
project = 'My Project'
copyright = '2021, My Team'
author = 'My Author'
version = '0.1'
release = version

source_suffix = ['.rst']

extensions = []

exclude_patterns = [
    'build'
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme_path = ['..']
html_theme = 'quaker_theme'
html_style = 'css/bulma.min.css'



"""
Utility file containing useful functionality for throughout the project.
"""

import re
from pathlib import Path

# From https://github.com/sphinx-doc/sphinx/blob/4.x/sphinx/util/nodes.py
_explicit_title_re = re.compile(r'^(.+?)\s*(?<!\x00)<([^<]*?)>$', re.DOTALL)

# From Docutils:
# https://github.com/docutils-mirror/docutils/blob/master/docutils/nodes.py
_non_id_chars = re.compile('[^a-z0-9/]+')
_non_id_at_ends = re.compile('^[-0-9]+|-+$')


def link_explicit(link: str):
    """Check whether the link format is a 'explicit link'.

    Explicit links are of the format:
    Link Title <link_ref>
    If this is the case, return tuple (Title, Reference).
    Otherwise return None.
    """
    res = _explicit_title_re.match(link)
    if res:
        title = res.group(1)
        ref = res.group(2)
        return (title, ref)

    return None


# https://stackoverflow.com/questions/4984647/accessing-dict-keys-like-an-attribute
class Config(dict):
    """
    Dictionary that can be accessed using attributes
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize Config
        """
        super(Config, self).__init__(*args, **kwargs)
        self.__dict__ = self


def make_id(ref: str) -> str:
    """Convert a string to an id format."""
    ref_list = ref.split('#')

    for i, ref in enumerate(ref_list):
        ref = str(Path(ref).with_suffix(''))
        ref = ref.lower()
        ref = _non_id_chars.sub('-', ' '.join(ref.split()))
        ref = _non_id_at_ends.sub('', ref)
        ref_list[i] = ref

    return '#'.join(ref_list)



from setuptools import setup, find_packages

# From https://github.com/pypa/pip/issues/7953#issuecomment-645133255
import site
import sys
site.ENABLE_USER_SITE = '--user' in sys.argv[1:]


setup(
    name = "quaker",
    version = "0.5.17",
    description = "Static Documentation Generator",
    long_description = open("README.md").read(),
    long_description_content_type = "text/markdown",
    entry_points = {
        'console_scripts': [
            'quaker=quaker:main',
        ],
    },
    url = "https://www.quakerdocs.nl/",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = ">=3.8",
    packages = find_packages("src"),
    package_dir = {"": "src"},
    include_package_data = True,
    package_data = {"quaker_lib": ["static/*", "wasm/*", "quickstart/*"]},
    install_requires = [
        "Pygments==2.9.0",
        "docutils==0.17.1",
        "Jinja2==2.10.3"
    ]
)



"""
Main entrypoint of the program.
Call this script to invoke the generation of a static document/website.
"""
import argparse
from pathlib import Path

from quaker_lib.main import Main


def main():
    """
    Main entrypoint of the program.
    """
    arg_parser = argparse.ArgumentParser(description='QuakerDocs')
    arg_parser.add_argument('source_path', type=Path,
                            help='The directory containing the RST files.')
    arg_parser.add_argument('-d', type=Path, dest='build_path',
                            default='build',
                            help='The directory to write the output.')
    arg_parser.add_argument('-b', type=str, dest='builder', default="html",
                            help='Builder used for the generator.')
    arg_parser.add_argument('--init', dest='init', action='store_true',
                            help="Initializes an empty project in the \
                                  specified source_path.")
    arg_parser.set_defaults(init=False)

    args = arg_parser.parse_args()

    print("Running QuakerDocs")
    main = Main(args.source_path, args.build_path, args.builder)

    if args.init:
        main.init_empty_project()
        return 0

    if not Path(args.source_path).is_dir():
        print("Error: not a directory")
        arg_parser.print_help()
        return 1
    if not Path(args.source_path).exists():
        print("Error: directory not found")
        arg_parser.print_help()
        return 1

    main.generate()
    return 0


if __name__ == "__main__":
    main()



"""Module that implements a directive to retrive metadata from a file.

This is a small directive and associated node that reads metadata from a file,
Rst or Md.

Notes
-----
The metadata variables should be of the form:
    name = value
    ...

Attributes
----------
default_fields : dictionary
    A dictionary of default values for metadata entries.
false_values : set
    A set of string values that represent a false boolean

"""

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives

# The metadata fields and their default values.
default_fields = {'priority': 1.0,
                  'ignore': False}

# The values which are interpreted as false.
false_values = {'false', 'f', '0', 'no', 'n'}


class metadata(nodes.Inline, nodes.Element):
    """
    Class for the metadata node
    """
    def __init__(self, fields):
        """Create nodes to contain the metadata."""
        nodes.Element.__init__(self)
        for key, value in fields.items():
            setattr(self, key, value)


class MetadataDirective(Directive):
    """A directive to get the metadata from the files.

    Attributes
    ----------
    has_content : bool
        Tell docutils that this directive uses it's content.

    """

    has_content = True

    def run(self):
        """ Run method that docutils uses on directive encounter. """

        fields = default_fields.copy()

        # Get the variables on each line
        for line in self.content:
            # Split on =
            var = line.replace(' ', '').split('=')

            # Check that the syntax was used correctly.
            if len(var) != 2:
                print(f'Warning: invalid metadata: {line}')
                print(f'\t At file {self.state.document.settings.page.path}')
                continue

            name, value = var

            # Parse the value and update the field.
            try:
                # Check if the default type is a bool and parse accordingly.
                if isinstance(fields[name], bool):
                    fields[name] = value.lower() not in false_values
                # Else use the default type to cast the string.
                else:
                    fields[name] = type(fields[name])(value)
            except (ValueError, KeyError):
                print(f'Warning: invalid metadata: {line}')
                print(f'\t At file {self.state.document.settings.page.path}')

        return [metadata(fields)]


def get_metadata(doctree):
    """ Helper function to get the metadata from a doctree.

    Gets the metadata of a parsed doctree, or creates the default metadata
    if it does not exist.

    Returns
    -------
     : metadata
        The node containing the metadata

    """
    try:
        return next(iter(doctree.traverse(lambda n: isinstance(n, metadata))))
    except StopIteration:
        return metadata(default_fields)


def setup():
    """ Add the metadata directives. """
    directives.register_directive('metadata', MetadataDirective)
